"""
tests module.

=========

Provides:
    1. Websocket and HTTP server methods tests.

How to use the documentation
----------------------------
Documentation is available in one form: docstrings provided
with the code

Copyright (c) 2016, Edgar A. Margffoy.
MIT, see LICENSE for more details.
"""
