html {
   margin: 0px;
   height: 100%;
   width: 100%;
}

body {
   margin: 0px;
   min-height: 100%;
   width: 100%;
}


h1 {
    text-align: center;
}

#terminal-container {
    height:100vh;
    margin:     0px;
    padding:    0px;
    z-index: 255;
}

#terminal-container .terminal {
    background-color: #000;
    color: #fafafa;
    padding: 0px;
    font-family: 'Ubuntu Mono', monospace;
    height:100%;
    font-size: 1em;
    line-height: 1 !important;
}


