# -*- coding: utf-8 -*-

"""
rest module.

=========

Provides:
    1. Asynchronous execution of JSON services

How to use the documentation
----------------------------
Documentation is available in one form: docstrings provided
with the code

Copyright (c) 2016, Edgar A. Margffoy.
MIT, see LICENSE for more details.
"""
import os
import sys
import rest.term_rest

__version__ = '1.0.0'
